"""
@copyright: 2007 Fusion Dealership Systems (Pty) Ltd. See LICENSE for details.
"""


class UnknownHashAlgorithm(ValueError):
    """
    An unknown hash algorithm was specified.
    """
    def __init__(self, algo):
        ValueError.__init__(self, algo)
        self.algo = algo



class CorruptObject(RuntimeError):
    """
    An object's contents did not have the expected digest.
    """



class NonexistentObject(ValueError):
    """
    The specified object does not exist.
    """
    def __init__(self, objectId):
        ValueError.__init__(self, objectId)
        self.objectId = objectId



class DigestMismatch(ValueError):
    """
    The content did not match the expected digest.

    @ivar expected: The expected digest.
    @ivar actual: The actual digest.
    """
    def __init__(self, expected, actual):
        self.expected = expected
        self.actual = actual


    def __repr__(self):
        return '<DigestMismatch expected=%r actual=%r>' % (
            self.expected, self.actual)


    def __str__(self):
        return 'Expected digest %r but got digest %r' % (
            self.expected, self.actual)



class APIError(RuntimeError):
    """
    A client's interaction with Entropy was interrupted by an error.
    """
    def __init__(self, message, code):
        """
        @type  code: L{int}
        @param code: Error code.
        """
        RuntimeError.__init__(self, message)
        self.code = code



class IrreparableError(RuntimeError):
    """
    An inconsistency was detected that cannot be repaired automatically.
    """
    def __init__(self, objectId):
        RuntimeError.__init__(self, objectId)
        self.objectId = objectId



class NoGoodCopies(IrreparableError):
    """
    No copies of an object could be found which match the content digest.
    """



class UnexpectedDigest(IrreparableError):
    """
    The metadata of object had a different digest when retrieved than expected.

    This exception signals a I{metadata} inconsistency; it does not refer to
    the actual object contents. This likely means that a backend is
    malfunctioning.
    """



"""
@copyright: 2007-2014 Quotemaster cc. See LICENSE for details.
"""
from twisted.trial.unittest import TestCase
from twisted.cred.portal import IRealm

from axiom.store import Store
from axiom.userbase import LoginSystem
from axiom.dependency import installOn

from xmantissa.ixmantissa import IOfferingTechnician
from xmantissa.offering import getOfferings

from entropy.util import getAppStore, deferred

class GetAppStoreTests(TestCase):
    """
    Tests for L{getAppStore}.
    """
    def setUp(self):
        self.siteStore = Store(self.mktemp())
        installOn(LoginSystem(store=self.siteStore), self.siteStore)

        for offering in getOfferings():
            if offering.name == u'Entropy':
                break
        else:
            raise RuntimeError(u'Could not find Entropy offering')

        IOfferingTechnician(self.siteStore).installOffering(offering)

    def test_getAppStore(self):
        """
        L{getAppStore} should return the Entropy application store.
        """
        appStore = IRealm(self.siteStore).accountByAddress(u'Entropy', None).avatars.open()
        self.assertIdentical(appStore, getAppStore(self.siteStore))


@deferred
def testfn(arg):
    """
    Test function for L{deferred} decorator.

    Raises L{ValueError} if 42 is passed, otherwise returns whatever was
    passed.
    """
    if arg == 42:
        raise ValueError('Oh noes')
    return arg


class DeferredTests(TestCase):
    """
    Tests for L{deferred} decorator.
    """
    def test_success(self):
        """
        Returning a value should result in a Deferred that callbacks with that
        value.
        """
        d = testfn(50)
        return d.addCallback(lambda result: self.assertEquals(result, 50))

    def test_failure(self):
        """
        Raising an exception should result in a Deferred that errbacks with
        that exception.
        """
        d = self.assertFailure(testfn(42), ValueError)
        return d.addCallback(lambda e: self.assertIn('Oh noes', str(e)))



"""
@copyright: 2007-2014 Quotemaster cc. See LICENSE for details.
"""
from twisted.trial.unittest import TestCase

from entropy.errors import UnknownHashAlgorithm, DigestMismatch

class ExceptionTests(TestCase):
    """
    Tests for exception classes.
    """
    def test_unknownHashAlgorithm(self):
        """
        Instantiating L{UnknownHashAlgorithm) correctly sets its attributes.
        """
        e = UnknownHashAlgorithm('algo')
        self.assertEquals(e.algo, 'algo')



class DigestMismatchTests(TestCase):
    """
    Tests for DigestMismatch.
    """
    def setUp(self):
        self.e = DigestMismatch('foo', 'bar')


    def test_str(self):
        """
        Verify the __str__ implementation.
        """
        self.assertEquals(
            "Expected digest 'foo' but got digest 'bar'",
            str(self.e))


    def test_repr(self):
        """
        Verify the __repr__ implementation.
        """
        self.assertEquals(
            "<DigestMismatch expected='foo' actual='bar'>",
            repr(self.e))



"""
@copyright: 2007-2014 Quotemaster cc. See LICENSE for details.
"""
from twisted.internet.defer import succeed
from twisted.python.failure import Failure
from twisted.web.client import ResponseDone
from twisted.web.http_headers import Headers
from twisted.web.test.test_agent import DummyResponse as _TwistedDummyResponse



class DummyResponse(_TwistedDummyResponse):
    """
    Dummy agent response.
    """
    def respond(self, data):
        """
        Deliver some data to the underlying protocol and close it.
        """
        self.protocol.dataReceived(data)
        self.protocol.connectionLost(Failure(ResponseDone()))



class DummyAgent(object):
    """
    Dummy L{IAgent} that only uses L{DummyResponse}s.
    """
    def __init__(self):
        self.responses = []


    def request(self, method, uri, headers=None, bodyProducer=None):
        if headers is None:
            headers = Headers({})
        response = DummyResponse()
        response.code = 200
        response.args = (method, uri, headers, bodyProducer)
        self.responses.append(response)
        return succeed(response)



"""
@copyright: 2007-2014 Quotemaster cc. See LICENSE for details.
"""
from zope.interface import implements

from epsilon.structlike import record

from twisted.internet import defer
from twisted.python.util import mergeFunctionMetadata

from xmantissa.offering import InstalledOffering

from entropy.ientropy import IContentObject


def getAppStore(siteStore):
    """
    Retrieve the Entropy app store.
    """
    offering = siteStore.findUnique(
        InstalledOffering,
        InstalledOffering.offeringName == u'Entropy')
    appStore = offering.application.open()
    return appStore


def deferred(f):
    def _wrapper(*a, **kw):
        return defer.execute(f, *a, **kw)
    return mergeFunctionMetadata(f, _wrapper)



class MemoryObject(record('content hash contentDigest contentType created '
                          'metadata', metadata={})):
    """
    In-memory implementation of L{IContentObject}.

    This is primarily useful for objects retrieved from a remote system, that
    need to be temporarily held in memory.
    """
    implements(IContentObject)


    @property
    def objectId(self):
        return u'%s:%s' % (self.hash, self.contentDigest)


    def getContent(self):
        return defer.succeed(self.content)



"""
@copyright: 2007-2014 Quotemaster cc. See LICENSE for details.
"""
import hashlib
from base64 import b64encode
from epsilon.extime import Time
from StringIO import StringIO
from twisted.internet import reactor
from twisted.python.urlpath import URLPath
from twisted.web import http
from twisted.web.client import Agent, readBody, FileBodyProducer
from twisted.web.http_headers import Headers

from entropy.errors import APIError
from entropy.util import MemoryObject



class Endpoint(object):
    """
    Entropy client endpoint.
    """
    def __init__(self, uri, agent=None):
        """
        @type  uri: L{unicode}
        @param uri: Entropy endpoint URI, for example:
            C{http://example.com/entropy/}.

        @type  agent: L{twisted.web.iweb.IAgent}
        @param agent: Twisted Web agent.
        """
        self.uri = URLPath.fromString(uri)
        if agent is None:
            agent = Agent(reactor)
        self._agent = agent


    def _parseResponse(self, response):
        """
        Parse an Entropy HTTP response.
        """
        def _checkResult(result):
            if response.code >= 400:
                raise APIError(result, response.code)
            return result, response

        d = readBody(response)
        d.addCallback(_checkResult)
        return d


    def store(self, content, contentType, metadata={}, created=None):
        """
        Store an object in an Entropy endpoint.

        @type  content: L{str}
        @param content: Object data.

        @type  contentType: L{unicode}
        @param contentType: MIME type of C{content}.

        @param metadata: Object metadata.
        @type  metadata: C{dict} of C{unicode}:C{unicode}

        @type  created: L{epsilon.extime.Time} or C{None}
        @param created: Creation timestamp; defaults to the current time.

        @raises L{APIError}: If there is an error storing the object in the
            Entropy endpoint.

        @rtype: L{Deferred} firing with L{unicode}
        @return: Object identifier.
        """
        if isinstance(contentType, unicode):
            contentType = contentType.encode('ascii')
        digest = hashlib.md5(content).digest()
        headers = Headers({
            'Content-Type': [contentType],
            'Content-MD5': [b64encode(digest)]})
        bodyProducer = FileBodyProducer(StringIO(content))
        d = self._agent.request(
            'PUT', str(self.uri.child('new')), headers, bodyProducer)
        d.addCallback(self._parseResponse)
        d.addCallback(lambda (result, response): result.decode('utf-8'))
        return d


    def get(self, objectId):
        """
        Retrieve an object from an Entropy endpoint.

        @type  objectId: L{unicode}
        @param objectId: Stored object identifier.

        @raises L{APIError}: If there is an error retrieving the object from
            the Entropy endpoint.

        @rtype: L{Deferred} firing with L{IContentObject}
        @return: Content object.
        """
        def _makeContentObject((data, response)):
            hash, contentDigest = objectId.split(u':', 1)
            contentType = response.headers.getRawHeaders(
                'Content-Type', ['application/octet-stream'])[0].decode('ascii')
            # XXX: Actually get the real creation time
            return MemoryObject(
                content=data,
                hash=hash,
                contentDigest=contentDigest,
                contentType=contentType,
                metadata={},
                created=Time())

        if not isinstance(objectId, unicode):
            objectId = objectId.decode('ascii')
        d = self._agent.request(
            'GET', str(self.uri.child(objectId.encode('ascii'))))
        d.addCallback(self._parseResponse)
        d.addCallback(_makeContentObject)
        return d


    def exists(self, objectId):
        """
        Determine if the specific Entropy object exists.

        @type  objectId: L{unicode}
        @param objectId: Stored object identifier.

        @raises L{APIError}: If there is an error retrieving the object from
            the Entropy endpoint.

        @rtype: L{Deferred} firing with L{bool}
        @return: Object exists?
        """
        def _checkNotFound(f):
            f.trap(APIError)
            if f.value.code == http.NOT_FOUND:
                return False
            return f

        d = self._agent.request(
            'HEAD', str(self.uri.child(objectId.encode('ascii'))))
        d.addCallback(self._parseResponse)
        d.addCallbacks(lambda ignored: True, _checkNotFound)
        return d



__all__ = ['Endpoint']



"""
Dropin with Entropy offering.
"""
from xmantissa.offering import Offering

from entropy.web import SimpleSiteFactory
from entropy.store import ContentResource


plugin = Offering(
    name=u'Entropy',
    description=u'A persistent, immutable object store.',
    siteRequirements=[(None, SimpleSiteFactory)],
    appPowerups=[ContentResource],
    installablePowerups=[],
    loginInterfaces=[],
    themes=[])



"""
@copyright: 2007 Fusion Dealership Systems (Pty) Ltd. See LICENSE for details.

Object data store.

This service acts as a cache / access point for a backend object store;
currently Amazon S3 is used as the backend store, but the architecture should
be flexible enough to allow other possibilities. The system is designed to
handle objects in an immutable fashion; once an object is created, it exists in
perpetuity, and the contents will never change.

The service's functionality is two-fold; firstly, it handles requests for
retrieval of objects, servicing them from the local cache, fetching them from a
neighbour cache, or retrieving them from the backend store. Secondly, it
handles requests for storage of a new object; the object is first cached
locally to ensure local view consistency, and then queued for backend storage
in a reliable fashion.
"""
import hashlib
from datetime import timedelta
from itertools import chain

from axiom.attributes import (
    AND, inmemory, integer, path, reference, text, timestamp)
from axiom.dependency import dependsOn
from axiom.iaxiom import IScheduler
from axiom.item import Item, transacted
from epsilon.extime import Time
from nevow.inevow import IRequest, IResource
from nevow.rend import NotFound
from nevow.static import File
from twisted.application.service import IService, Service
from twisted.internet.defer import execute, fail, gatherResults, succeed
from twisted.internet.task import cooperate
from twisted.internet.threads import deferToThread
from twisted.logger import Logger
from twisted.python.components import registerAdapter
from twisted.python.filepath import FilePath
from twisted.web import http
from zope.interface import implements

from entropy.client import Endpoint
from entropy.errors import (
    APIError, CorruptObject, DigestMismatch, NoGoodCopies, NonexistentObject,
    UnexpectedDigest)
from entropy.hash import getHash
from entropy.ientropy import (
    IBackendStore, IContentObject, IContentStore, IMigration,
    IMigrationManager, ISiblingStore, IUploadScheduler)
from entropy.util import deferred



class ImmutableObject(Item):
    """
    An immutable object.

    Immutable objects are addressed by content hash, and consist of the object
    data as a binary blob, and object key/value metadata pairs.
    """
    implements(IContentObject)

    hash = text(allowNone=False)
    contentDigest = text(allowNone=False, indexed=True)
    content = path(allowNone=False)
    contentType = text(allowNone=False)
    created = timestamp(allowNone=False, defaultFactory=lambda: Time())

    _deferToThreadPool = inmemory()

    def activate(self):
        self._deferToThreadPool = execute


    @property
    def metadata(self):
        return {}


    @property
    def objectId(self):
        return u'%s:%s' % (self.hash, self.contentDigest)


    def _getDigest(self):
        fp = self.content.open()
        try:
            h = getHash(self.hash)(fp.read())
            return unicode(h.hexdigest(), 'ascii')
        finally:
            fp.close()


    def verify(self):
        digest = self._getDigest()
        if self.contentDigest != digest:
            raise CorruptObject(
                'expected: %r actual: %r' % (self.contentDigest, digest))


    def getContent(self):
        return self._deferToThreadPool(self.content.getContent)


def objectResource(obj):
    """
    Adapt L{ImmutableObject) to L{IResource}.
    """
    res = File(obj.content.path)
    res.type = obj.contentType.encode('ascii')
    res.encoding = None
    return res

registerAdapter(objectResource, ImmutableObject, IResource)



class PendingMigration(Item):
    """
    An item that tracks the state the migration of an individual object.

    Once a migration process decides to migrate a particular object, an
    instance of this item will be created to track the migration of the object,
    and will only be deleted once the object has been successfully migrated.
    """
    parent = reference(
        allowNone=False,
        doc="The migration to which this object belongs.")
    obj = reference(
        allowNone=False, reftype=ImmutableObject,
        doc="The object being migrated.")
    lastFailure = text(
        doc="A description of the last failed migration attempt, if any.")


    def _verify(self):
        """
        Verify an object.

        The contents of the object in every store is verified against the
        content digest; any missing or corrupt objects are replaced with a
        non-corrupt version, if possible.

        @raise NoGoodCopies: If no good copy could be found to repair from.

        @raise UnexpectedDigest: If a backend returned a different object to
        the one requested.
        """
        def getContent(obj):
            return obj.getContent().addCallback(lambda content: (obj, content))

        def handleMissing(f):
            f.trap(NonexistentObject)
            return None, None

        def gotContents(cs):
            hashFunc = getHash(self.obj.hash)
            expected = self.obj.contentDigest
            corrupt = []
            goodObj = None
            goodContent = None
            for backend, (obj, content) in zip(backends, cs):
                if content is None:
                    corrupt.append(backend)
                    continue
                if obj.contentDigest != expected:
                    log.error(
                        'Unexpected content digest mismatch! '
                        '{left!s} != {right!s} in {backend!r}',
                        left=expected,
                        right=obj.contentDigest,
                        backend=backend)
                    raise UnexpectedDigest(objectId)
                if unicode(hashFunc(content).hexdigest(), 'ascii') == expected:
                    if goodObj is None:
                        goodObj = obj
                        goodContent = content
                else:
                    corrupt.append(backend)
                    path = FilePath('corrupt').temporarySibling()
                    log.debug(
                        'Saving corrupt content for {objectId!s} from '
                        '{backend!r} in {path!s}.',
                        objectId=objectId,
                        backend=backend,
                        path=path)
                    path.setContent(content)
            if goodObj is None:
                log.error(
                    'All copies of {objectId!s} are corrupt, unable to repair!',
                    objectId=objectId)
                raise NoGoodCopies(objectId)
            else:
                ds = []
                for backend in corrupt:
                    log.info(
                        'Repairing {objectId!s} in {backend!r}',
                        objectId=self.obj.objectId, backend=backend)
                    ds.append(backend.storeObject(
                        content=goodContent,
                        contentType=goodObj.contentType,
                        metadata=goodObj.metadata,
                        created=goodObj.created,
                        objectId=objectId))
                return gatherResults(ds, consumeErrors=True)

        objectId = self.obj.objectId
        backends = [self.parent.source]
        backends.extend(self.store.powerupsFor(ISiblingStore))
        backends.extend(self.store.powerupsFor(IBackendStore))
        contents = [succeed(self.obj).addCallback(getContent)] + [
            backend.getObject(objectId).addCallbacks(
                getContent, handleMissing)
            for backend in backends[1:]]
        return gatherResults(contents, consumeErrors=True).addCallback(
            gotContents)


    def _migrate(self):
        """
        Migrate an object from one store to another.
        """
        if self.parent.destination is None:
            # This is a "verification" migration
            return self._verify()

        d = self.obj.getContent()
        d.addCallback(
            lambda content: self.parent.destination.storeObject(
                content=content,
                contentType=self.obj.contentType,
                metadata=self.obj.metadata,
                created=self.obj.created,
                objectId=self.obj.objectId))
        return d


    def attemptMigration(self):
        """
        Perform one attempt at migration of the object being tracked.

        If the migration is successful, this item will be deleted; otherwise,
        the failure will be stored in the C{lastFailure} attribute.

        @rtype: Deferred<None>
        """
        def _cb(ign):
            self.deleteFromStore()

        def _eb(f):
            log.failure(
                'Error during migration of {objectId!r} from {source!r} to {destination!r}',
                failure=f,
                objectId=self.obj.objectId,
                source=self.parent.source,
                destination=self.parent.destination)
            self.lastFailure = unicode(
                f.getTraceback(), 'ascii', errors='replace')

        return self._migrate().addCallbacks(_cb, _eb)



class LocalStoreMigration(Item):
    """
    Migration from local content store.
    """
    implements(IMigration)
    powerupInterfaces = [IMigration]

    source = reference(
        allowNone=False,
        doc="The content store that is the source of this migration")
    destination = reference(
        allowNone=True,
        doc="""
        The content store that is the destination of this migration. If
        C{None}, this is a "validation" migration.
        """)
    start = integer(allowNone=False, doc="Starting storeID")
    current = integer(allowNone=False, doc="Most recent storeID migrated")
    end = integer(allowNone=False, doc="Ending storeID")

    concurrency = 10

    _running = inmemory()

    def activate(self):
        self._running = False


    @transacted
    def _nextObject(self):
        """
        Obtain the next object for which migration should be attempted.
        """
        obj = self.store.findFirst(
            ImmutableObject,
            AND(ImmutableObject.storeID > self.current,
                ImmutableObject.storeID <= self.end),
            sort=ImmutableObject.storeID.asc)
        if obj is None:
            return None
        self.current = obj.storeID
        return PendingMigration(store=self.store, parent=self, obj=obj)


    # IMigration

    def run(self):
        """
        Perform the migration.
        """
        if self._running:
            return
        self._running = True

        def _done(ign):
            self._running = False

        migrations = chain(
            self.store.query(
                PendingMigration, PendingMigration.parent == self),
            iter(self._nextObject, None))
        it = (m.attemptMigration() for m in migrations)
        tasks = [cooperate(it) for _ in xrange(self.concurrency)]
        d = gatherResults([task.whenDone() for task in tasks])
        d.addCallback(_done)
        return d



class ContentStore(Item):
    """
    Manager for stored objects.
    """
    implements(IContentStore)
    powerupInterfaces = [IContentStore]

    hash = text(allowNone=False, default=u'sha256')

    def _deferToThreadPool(self, f, *a, **kw):
        return deferToThread(f, *a, **kw)


    @transacted
    def _storeObject(self, content, contentType, metadata={}, created=None):
        """
        Do the actual work of synchronously storing the object.
        """
        if metadata != {}:
            raise NotImplementedError('metadata not yet supported')

        contentDigest = getHash(self.hash)(content).hexdigest()
        contentDigest = unicode(contentDigest, 'ascii')

        if created is None:
            created = Time()

        obj = self.store.findUnique(
            ImmutableObject,
            AND(ImmutableObject.hash == self.hash,
                ImmutableObject.contentDigest == contentDigest),
            default=None)
        if obj is None:
            bucket = contentDigest[:3]
            contentFile = self.store.newFile(
                'objects', 'immutable', bucket,
                '%s:%s' % (self.hash, contentDigest))
            contentFile.write(content)
            contentFile.close().addErrback(
                lambda f: log.failure(
                    'Error writing object to {name!r}:',
                    f, name=contentFile.name))

            obj = ImmutableObject(store=self.store,
                                  contentDigest=contentDigest,
                                  hash=self.hash,
                                  content=contentFile.finalpath,
                                  contentType=contentType,
                                  created=created)
            obj._deferToThreadPool = self._deferToThreadPool
        else:
            obj.contentType = contentType
            obj.created = created
            obj.content.setContent(content)
            obj._deferToThreadPool = self._deferToThreadPool

        scheduler = IUploadScheduler(self.store, None)
        for backend in self.store.powerupsFor(IBackendStore):
            if scheduler is None:
                raise RuntimeError('No upload scheduler configured')
            scheduler.scheduleUpload(obj.objectId, backend)

        return obj


    def importObject(self, obj):
        """
        Import an object from elsewhere.

        @param obj: the object to import.
        @type obj: ImmutableObject
        """
        return obj.getContent().addCallback(
            lambda content: self._storeObject(
                content,
                obj.contentType,
                obj.metadata,
                obj.created))


    @transacted
    def getSiblingObject(self, objectId):
        """
        Import an object from a sibling store.

        @returns: the local imported object.
        @type obj: ImmutableObject
        """
        siblings = list(self.store.powerupsFor(ISiblingStore))
        siblings.extend(self.store.powerupsFor(IBackendStore))
        siblings = iter(siblings)

        def _eb(f):
            f.trap(NonexistentObject)
            try:
                remoteStore = siblings.next()
            except StopIteration:
                raise NonexistentObject(objectId)

            d = remoteStore.getObject(objectId)
            d.addCallbacks(self.importObject, _eb)
            return d

        return self.getObject(objectId).addErrback(_eb)


    # IContentStore

    @deferred
    def storeObject(self, content, contentType, metadata={}, created=None,
                    objectId=None):
        obj = self._storeObject(content, contentType, metadata, created)
        return obj.objectId


    @deferred
    @transacted
    def getObject(self, objectId):
        hash, contentDigest = objectId.split(u':', 1)
        obj = self.store.findUnique(
            ImmutableObject,
            AND(ImmutableObject.hash == hash,
                ImmutableObject.contentDigest == contentDigest),
            default=None)
        if obj is None:
            raise NonexistentObject(objectId)
        obj._deferToThreadPool = self._deferToThreadPool
        return obj


    @transacted
    def migrateTo(self, destination):
        latestObject = self.store.findFirst(
            ImmutableObject, sort=ImmutableObject.storeID.desc)
        return LocalStoreMigration(
            store=self.store,
            source=self,
            destination=destination,
            start=0,
            current=-1,
            end=latestObject.storeID)



class ObjectCreator(object):
    """
    Resource for storing new objects.

    @ivar contentStore: The {IContentStore} provider to create objects in.
    """
    implements(IResource)

    def __init__(self, contentStore):
        self.contentStore = contentStore


    # IResource
    def renderHTTP(self, ctx):
        req = IRequest(ctx)
        if req.method == 'GET':
            req.setHeader('Content-Type', 'text/plain')
            return 'PUT data here to create an object.'
        elif req.method == 'PUT':
            return self.handlePUT(req)
        else:
            req.setResponseCode(http.NOT_ALLOWED)
            req.setHeader('Content-Type', 'text/plain')
            return 'Method not allowed'


    def handlePUT(self, req):
        data = req.content.read()
        contentType = unicode(
            req.getHeader('Content-Type') or 'application/octet-stream',
            'ascii')

        contentMD5 = req.getHeader('Content-MD5')
        if contentMD5 is not None:
            expectedHash = contentMD5.decode('base64')
            actualHash = hashlib.md5(data).digest()
            if expectedHash != actualHash:
                raise DigestMismatch(expectedHash, actualHash)

        def _cb(objectId):
            req.setHeader('Content-Type', 'text/plain')
            objectId = objectId.encode('ascii')
            return objectId

        d = self.contentStore.storeObject(data, contentType)
        return d.addCallback(_cb)



class ContentResource(Item):
    """
    Resource for accessing the content store.
    """
    implements(IResource)
    powerupInterfaces = [IResource]

    addSlash = inmemory()

    contentStore = dependsOn(ContentStore)

    def getObject(self, name):
        def _notFound(f):
            f.trap(NonexistentObject)
            return None
        return self.contentStore.getSiblingObject(name).addErrback(_notFound)


    def childFactory(self, name):
        """
        Hook up children.

        / is the root, nothing to see her.

        /new is how new objects are stored.

        /<objectId> is where existing objects are retrieved.
        """
        if name == '':
            return self
        elif name == 'new':
            return ObjectCreator(self.contentStore)
        else:
            return self.getObject(name)
        return None


    # IResource
    def renderHTTP(self, ctx):
        """
        Nothing to see here.
        """
        return 'Entropy'


    def locateChild(self, ctx, segments):
        """
        Dispatch to L{childFactory}.
        """
        if len(segments) >= 1:
            res = self.childFactory(segments[0])
            if res is not None:
                return res, segments[1:]
        return NotFound



class RemoteEntropyStore(Item):
    """
    IContentStore implementation for remote Entropy services.
    """
    implements(IContentStore)

    entropyURI = text(allowNone=False,
                      doc="""The URI of the Entropy service in use.""")

    _endpoint = inmemory()


    def activate(self):
        self._endpoint = Endpoint(uri=self.entropyURI)


    # IContentStore

    def storeObject(self, content, contentType, metadata={}, created=None,
                    objectId=None):
        return self._endpoint.store(
            content=content,
            contentType=contentType,
            metadata=metadata,
            created=created)


    def getObject(self, objectId):
        def _checkError(f):
            f.trap(APIError)
            if f.value.code == http.NOT_FOUND:
                return fail(NonexistentObject(objectId))
            return f

        d = self._endpoint.get(objectId)
        d.addErrback(_checkError)
        return d



class _PendingUpload(Item):
    """
    Marker for a pending upload to a backend store.
    """
    objectId = text(allowNone=False)
    backend = reference(allowNone=False) # reftype=IBackendStore
    scheduled = timestamp(
        indexed=True, allowNone=False, defaultFactory=lambda: Time())


    def _nextAttempt(self):
        """
        Determine the time to schedule the next attempt.
        """
        return Time() + timedelta(minutes=2)


    def run(self):
        self.attemptUpload()


    def attemptUpload(self):
        """
        Attempt an upload of an object to a backend store.

        If the upload fails, it will be rescheduled; if it succeeds, this item
        will be deleted.
        """
        def _uploadObject(obj):
            d = obj.getContent()
            d.addCallback(
                lambda content: self.backend.storeObject(
                    content,
                    obj.contentType,
                    obj.metadata,
                    obj.created,
                    objectId=self.objectId))
            return d

        def _reschedule(f):
            # We do this instead of returning a Time from attemptUpload,
            # because that can only be done synchronously.
            log.failure(
                'Error uploading object {objectId!r} to backend store {backend!r}',
                failure=f,
                objectId=self.objectId,
                backend=self.backend)
            self.scheduled = self._nextAttempt()
            self.schedule()

        d = succeed(None)
        d.addCallback(
            lambda ign: IContentStore(self.store).getObject(self.objectId))
        d.addCallback(_uploadObject)
        d.addCallbacks(lambda ign: self.deleteFromStore(), _reschedule)
        return d


    def schedule(self):
        IScheduler(self.store).schedule(self, self.scheduled)



class UploadScheduler(Item):
    """
    Schedule upload attempts for pending uploads.
    """
    implements(IUploadScheduler)
    powerupInterfaces = [IUploadScheduler]

    dummy = text()

    # IUploadScheduler

    def scheduleUpload(self, objectId, backend):
        upload = _PendingUpload(
            store=self.store,
            objectId=objectId,
            backend=backend)
        upload.schedule()



class MigrationManager(Item, Service):
    """
    Default migration manager implementation.
    """
    implements(IMigrationManager, IService)
    powerupInterfaces = [IMigrationManager, IService]

    dummy = integer()

    # IService
    parent = inmemory()
    name = inmemory()
    running = inmemory()

    def activate(self):
        self.parent = None
        self.name = None
        self.running = False


    def installed(self):
        """
        Callback invoked after this item has been installed on a store.

        This is used to set the service parent to the store's service object.
        """
        self.setServiceParent(self.store)


    def deleted(self):
        """
        Callback invoked after a transaction in which this item has been
        deleted is committed.

        This is used to remove this item from its service parent, if it has
        one.
        """
        if self.parent is not None:
            self.disownServiceParent()


    # IMigrationManager

    def migrate(self, source, destination):
        """
        Initiate a migration between two content stores.

        @see: L{entropy.ientropy.IMigrationManager.migrate}
        """
        migration = source.migrateTo(destination)
        self.store.powerUp(migration, IMigration)
        migration.run()
        return migration


    # IService

    def startService(self):
        self.running = True
        for migration in self.store.powerupsFor(IMigration):
            migration.run()



log = Logger()



"""
@copyright: 2007-2014 Quotemaster cc. See LICENSE for details.
"""
from entropy.errors import UnknownHashAlgorithm

from hashlib import sha256

_hashes = {
    u'sha256': sha256,
    }

def getHash(algo):
    try:
        return _hashes[algo]
    except KeyError:
        raise UnknownHashAlgorithm(algo)



"""
@copyright: 2007-2014 Quotemaster cc. See LICENSE for details.
"""
from twisted.trial.unittest import TestCase

from entropy.hash import getHash
from entropy.errors import UnknownHashAlgorithm


class HashingTests(TestCase):
    """
    Tests for L{entropy.hash}.
    """
    def test_sha256(self):
        """
        Retrieving the sha256 hash function succeeds.
        """
        getHash('sha256')
        self.assertTrue(callable(hash))


    def test_invalidHash(self):
        """
        Trying to retrieve an unknown hash function results in an
        L{UnknownHashAlgorithm} exception.
        """
        self.assertRaises(UnknownHashAlgorithm, getHash, '***DOESNOTEXIST***')



"""
@copyright: 2007-2014 Quotemaster cc. See LICENSE for details.
"""
from StringIO import StringIO
from twisted.trial.unittest import TestCase
from twisted.web import http
from twisted.web.http_headers import Headers

from entropy.client import Endpoint
from entropy.errors import APIError
from entropy.test.util import DummyAgent



class _FileConsumer(object):
    """
    Trivial L{IConsumer} that writes its input to another output.
    """
    def __init__(self, outputFile):
        self.outputFile = outputFile


    def write(self, data):
        self.outputFile.write(data)



class EndpointTests(TestCase):
    """
    Tests for L{entropy.client.Endpoint}.
    """
    def setUp(self):
        self.agent = DummyAgent()
        self.endpoint = Endpoint(u'http://example.com/entropy/', self.agent)


    def test_failure(self):
        """
        If Entropy returns a non-success code, L{Endpoint} raises L{APIError}.
        """
        d = self.endpoint.store('some_data', 'text/plain')
        response = self.agent.responses.pop()
        self.assertEqual([], self.agent.responses)
        response.code = http.BAD_REQUEST
        response.respond('Oops')
        f = self.failureResultOf(d, APIError)
        self.assertEqual(
            (http.BAD_REQUEST, 'Oops'),
            (f.value.code, f.value.message))


    def test_success(self):
        """
        Parse a successful Entropy response.
        """
        d = self.endpoint.store('some_data', 'text/plain')
        response = self.agent.responses.pop()
        self.assertEqual([], self.agent.responses)
        response.respond('an_id')
        self.assertEqual('an_id', self.successResultOf(d))


    def test_store(self):
        """
        Store an object in an Entropy endpoint.
        """
        d = self.endpoint.store('some_data', 'text/plain')
        response = self.agent.responses.pop()
        self.assertEqual([], self.agent.responses)
        self.assertEqual(
            ('PUT', 'http://example.com/entropy/new'),
            (response.args[0], response.args[1]))
        self.assertEqual(
            Headers({
                'Content-Type': ['text/plain'],
                'Content-MD5': ['DZJHy840q6SsqNXIh6DwpA==']}),
            response.args[2])
        response.respond('an_id')
        self.assertEqual('an_id', self.successResultOf(d))

        def _checkBody(result):
            self.assertEqual(
                'some_data',
                output.getvalue())
            return result

        output = StringIO()
        consumer = _FileConsumer(output)
        complete = response.args[3].startProducing(consumer)
        complete.addCallback(_checkBody)
        return complete


    def test_get(self):
        """
        Retrieve an existing Entropy object.
        """
        d = self.endpoint.get(u'sha256:an_id')
        response = self.agent.responses.pop()
        self.assertEqual([], self.agent.responses)
        response.headers.setRawHeaders('Content-Type', ['applicaton/pdf'])
        self.assertEqual(
            ('GET', 'http://example.com/entropy/sha256:an_id'),
            (response.args[0], response.args[1]))
        response.respond('some_data')
        obj = self.successResultOf(d)
        self.assertEqual(
            ('some_data', u'sha256', u'an_id', u'applicaton/pdf', {}),
            (obj.content, obj.hash, obj.contentDigest, obj.contentType,
             obj.metadata))


    def test_exists(self):
        """
        Determine if an Entropy object exists.
        """
        d = self.endpoint.exists(u'sha256:an_id')
        response = self.agent.responses.pop()
        self.assertEqual([], self.agent.responses)
        self.assertEqual(
            ('HEAD', 'http://example.com/entropy/sha256:an_id'),
            (response.args[0], response.args[1]))
        response.respond('')
        self.assertTrue(self.successResultOf(d))


    def test_doesNotExist(self):
        """
        Determine if an Entropy object does not exist.
        """
        d = self.endpoint.exists(u'sha256:an_id')
        response = self.agent.responses.pop()
        self.assertEqual([], self.agent.responses)
        response.code = http.NOT_FOUND
        self.assertEqual(
            ('HEAD', 'http://example.com/entropy/sha256:an_id'),
            (response.args[0], response.args[1]))
        response.respond('')
        self.assertFalse(self.successResultOf(d))



"""
@copyright: 2007-2014 Quotemaster cc. See LICENSE for details.
"""
from zope.interface import implements

from axiom.item import Item
from axiom.attributes import path

from nevow.inevow import IResource

from xmantissa.ixmantissa import IProtocolFactoryFactory
from xmantissa.web import AxiomSite

from entropy.util import getAppStore


# FIXME: This is a blatant rip from Mantissa's code, necessary in order to
# customize the root of the site, to avoid having guard wrapped around it.

class SimpleSiteFactory(Item):
    """
    Configuration object for a Mantissa HTTP server.
    """
    powerupInterfaces = [IProtocolFactoryFactory]
    implements(*powerupInterfaces)

    httpLog = path(default=None)

    def getFactory(self):
        """
        Create an L{AxiomSite} which supports authenticated and anonymous
        access.
        """
        logPath = self.httpLog and self.httpLog.path
        appStore = getAppStore(self.store)
        return AxiomSite(appStore, IResource(appStore), logPath=logPath)



"""
@copyright: 2007-2014 Quotemaster cc. See LICENSE for details.

Interface definitions for Entropy.
"""
from zope.interface import Interface, Attribute



class IContentObject(Interface):
    """
    Immutable content object.
    """
    hash = Attribute("""The hash function used to calculate the content digest.""")
    contentDigest = Attribute("""A digest of the object content.""")
    contentType = Attribute("""The MIME type describing the content of this object.""")
    created = Attribute("""Creation timestamp of this object.""")
    metadata = Attribute("""Object metadata.""")

    def getContent():
        """
        Get the data contained in this object.

        @rtype: C{str}
        """



class IContentStore(Interface):
    """
    Interface for storing and retrieving immutable content objects.
    """
    def storeObject(content, contentType, metadata={}, created=None):
        """
        Store an object.

        @param content: the data to store.
        @type content: C{str}

        @param contentType: the MIME type of the content.
        @type contentType: C{unicode}

        @param metadata: a dictionary of metadata entries.
        @type metadata: C{dict} of C{unicode}:C{unicode}

        @param created: the creation timestamp; defaults to the current time.
        @type created: L{epsilon.extime.Time} or C{None}

        @returns: the object identifier.
        @rtype: C{Deferred<unicode>}
        """


    def getObject(objectID):
        """
        Retrieve an object.

        @param objectId: the object identifier.
        @type objectId: C{unicode}
        @returns: the content object.
        @rtype: C{Deferred<IContentObject>}
        """


    def migrateTo(destination):
        """
        Initiate a migration to another content store.

        All objects present in this content store at the moment the migration
        is initiated MUST be replicated to the destination store before the
        migration is considered complete. Objects created after the migration
        is initiated MUST NOT be replicated.

        NOTE: This method is optional, as some storage backends may be unable
        to support enumerating all objects which is usually necessary to
        implement migration.

        @type  destination: L{IContentStore}
        @param destination: The destination store.

        @rtype: L{IMigration}
        @return: The migration powerup tracking the requested migration.

        @raise NotImplementedError: if this implementation does not support
            migration.
        """



class ISiblingStore(IContentStore):
    """
    Sibling content store.
    """



class IBackendStore(IContentStore):
    """
    Backend content store.
    """



class IUploadScheduler(Interface):
    """
    Manager of pending uploads.
    """
    def scheduleUpload(objectId, backend):
        """
        Notify the scheduler that an object needs to be uploaded to a backend.
        """



class IMigrationManager(Interface):
    """
    Manager for migrations from one content store to another.
    """
    def migrate(source, destination):
        """
        Initiate a migration between two content stores. Some content stores
        may not support migration, as some storage backends cannot support
        enumerating all stored objects.

        @type  source: L{IContentStore}
        @param source: The source content store; must support migration.

        @type  destination: L{IContentStore}
        @param destination: The destination store; does not need any special
            support for migration.

        @rtype: L{IMigration}
        @return: The migration powerup responsible for tracking the requested
            migration.
        """



class IMigration(Interface):
    """
    Powerup tracking a migration in progress.
    """
    def run():
        """
        Run this migration.

        If the migration is already running, this is a noop.
        """



import versioneer
from setuptools import find_packages, setup
setup(
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    name='entropy-store',
    maintainer='Entropy developers',
    description='A simple content-addressed immutable object store, with flexible backend support.',
    url='https://github.com/fusionapp/entropy',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Twisted',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.5',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Archiving'],
    packages=find_packages() + ['xmantissa.plugins'],
    install_requires=['Twisted[tls] >= 15.2.1',
                      'Epsilon >= 0.7.0',
                      'Axiom >= 0.7.4',
                      'Nevow >= 0.9.8',
                      'txAWS >= 0.2',
                      'Mantissa >= 0.8.0'])



"""
@copyright: 2011-2014 Quotemaster cc. See LICENSE for details.

Backend implementation using Amazon S3 for storage.
"""
from zope.interface import implements

from axiom.item import Item
from axiom.attributes import text

from txaws.service import AWSServiceRegion
from txaws.credentials import AWSCredentials
from txaws.s3.exception import S3Error

from entropy.ientropy import IContentStore
from entropy.util import MemoryObject
from entropy.errors import NonexistentObject


class S3Store(Item):
    """
    Content store using Amazon S3.
    """
    implements(IContentStore)

    accessKey = text(allowNone=False, doc="AWS access key.")
    secretKey = text(allowNone=False, doc="AWS secret key.")
    bucket = text(allowNone=False, doc="Name of S3 bucket used for storage.")

    def _getClient(self):
        """
        Build a txAWS S3 client using our stored credentials.
        """
        creds = AWSCredentials(
            access_key=self.accessKey.encode('utf-8'),
            secret_key=self.secretKey.encode('utf-8'))
        region = AWSServiceRegion(creds=creds)
        return region.get_s3_client()


    # IContentStore

    def storeObject(self, content, contentType, metadata={}, created=None,
                    objectId=None):
        if objectId is None:
            raise NotImplementedError('Must provide objectId')
        if metadata != {}:
            raise NotImplementedError('Metadata not supported')

        client = self._getClient()
        d = client.put_object(
            bucket=self.bucket.encode('utf-8'),
            object_name=objectId.encode('utf-8'),
            data=content,
            content_type=contentType.encode('utf-8'))
        d.addCallback(lambda ign: objectId)
        return d


    def getObject(self, objectId):
        hash, contentDigest = objectId.split(u':', 1)
        def _makeObject(content):
            headers = query.get_response_headers()
            return MemoryObject(
                content=content,
                hash=hash,
                contentDigest=contentDigest,
                contentType=unicode(headers['content-type'][0], 'utf-8'),
                created=None)

        def _eb(f):
            f.trap(S3Error)
            raise NonexistentObject(objectId)

        client = self._getClient()
        query = client.query_factory(
            action='GET', creds=client.creds, endpoint=client.endpoint,
            bucket=self.bucket.encode('utf-8'),
            object_name=objectId.encode('utf-8'))
        d = query.submit()
        d.addCallbacks(_makeObject, _eb)
        return d



"""
@copyright: 2007-2014 Quotemaster cc. See LICENSE for details.

Tests for L{entropy.store}.
"""
from datetime import timedelta
from StringIO import StringIO

from axiom.attributes import inmemory, integer
from axiom.dependency import installOn
from axiom.errors import ItemNotFound
from axiom.item import Item
from axiom.store import Store
from epsilon.extime import Time
from nevow.inevow import IResource
from nevow.static import File
from nevow.testutil import FakeRequest
from twisted.application.service import IService
from twisted.internet.defer import execute, fail, succeed
from twisted.python.components import proxyForInterface
from twisted.trial.unittest import TestCase
from twisted.web import http
from zope.interface import implementer

from entropy.client import Endpoint
from entropy.errors import (
    CorruptObject, NoGoodCopies, NonexistentObject, UnexpectedDigest)
from entropy.ientropy import (
    IBackendStore, IContentStore, IMigration, ISiblingStore, IUploadScheduler)
from entropy.store import (
    ContentStore, ImmutableObject, LocalStoreMigration, MigrationManager,
    ObjectCreator, PendingMigration, RemoteEntropyStore, _PendingUpload)
from entropy.test.util import DummyAgent
from entropy.util import MemoryObject



class RemoteEntropyStoreTests(TestCase):
    """
    Tests for L{entropy.store.RemoteEntropyStore}.
    """
    def setUp(self):
        self.uri = u'http://localhost:8080/'
        self.agent = DummyAgent()
        self.store = Store(self.mktemp())
        self.remoteEntropyStore = RemoteEntropyStore(
            store=self.store,
            entropyURI=self.uri)
        object.__setattr__(
            self.remoteEntropyStore,
            '_endpoint',
            Endpoint(uri=self.uri,
                     agent=self.agent))


    def test_nonexistentObject(self):
        """
        Retrieving a nonexistent object results in L{NonexistentObject}.
        """
        objectId = u'sha256:NOSUCHOBJECT'
        d = self.remoteEntropyStore.getObject(objectId)
        response = self.agent.responses.pop()
        self.assertEqual([], self.agent.responses)
        response.code = http.NOT_FOUND
        response.respond('Not found')
        f = self.failureResultOf(d, NonexistentObject)
        self.assertEqual(f.value.objectId, objectId)



class ContentStoreTests(TestCase):
    """
    Tests for L{ContentStore}.
    """
    def setUp(self):
        self.store = Store(self.mktemp())
        self.contentStore = ContentStore(store=self.store, hash=u'sha256')
        object.__setattr__(self.contentStore, '_deferToThreadPool', execute)


    def test_storeObject(self):
        """
        Test storing an object.
        """
        content = 'blahblah some data blahblah'
        contentType = u'application/octet-stream'
        expectedDigest = u'9aef0e119873bb0aab04e941d8f76daf21dedcd79e2024004766ee3b22ca9862'

        d = self.contentStore.storeObject(content, contentType)
        def _cb(oid):
            self.oid = oid
        d.addCallback(_cb)
        self.assertEquals(self.oid, u'sha256:' + expectedDigest)


    def test_metadata(self):
        """
        Attempting to store metadata results in an exception as this is not yet
        implemented.
        """
        d = self.contentStore.storeObject(
            'blah', 'blah', metadata={'blah': 'blah'})
        return self.assertFailure(d, NotImplementedError
            ).addCallback(lambda e: self.assertSubstring('metadata', str(e)))


    def test_getObject(self):
        """
        Test retrieving object.
        """
        obj = ImmutableObject(store=self.store,
                              hash=u'somehash',
                              contentDigest=u'quux',
                              content=self.store.newFilePath('foo'),
                              contentType=u'application/octet-stream')
        d = self.contentStore.getObject(u'somehash:quux')
        return d.addCallback(lambda obj2: self.assertIdentical(obj, obj2))


    def test_updateObject(self):
        """
        Storing an object that is already in the store just updates the content
        type and timestamp.
        """
        t1 = Time()
        t2 = t1 - timedelta(seconds=30)
        obj = self.contentStore._storeObject('blah',
                                             u'application/octet-stream',
                                             created=t1)
        obj2 = self.contentStore._storeObject('blah',
                                              u'text/plain',
                                              created=t2)
        self.assertIdentical(obj, obj2)
        self.assertEquals(obj.contentType, u'text/plain')
        self.assertEquals(obj.created, t2)

        self.contentStore._storeObject('blah', u'text/plain')

        self.assertTrue(obj.created > t2)


    def test_importObject(self):
        """
        Importing an object stores an equivalent object in the local store.
        """
        created = Time()

        obj1 = MemoryObject(hash=u'sha256',
                            contentDigest=u'9aef0e119873bb0aab04e941d8f76daf21dedcd79e2024004766ee3b22ca9862',
                            content=u'blahblah some data blahblah',
                            created=created,
                            contentType=u'application/octet-stream')
        obj2 = self.successResultOf(self.contentStore.importObject(obj1))
        self.assertEquals(obj1.objectId, obj2.objectId)
        self.assertEquals(obj1.created, obj2.created)
        self.assertEquals(obj1.contentType, obj2.contentType)
        self.assertEquals(
            self.successResultOf(obj1.getContent()),
            self.successResultOf(obj2.getContent()))


    def test_nonexistentObject(self):
        """
        Retrieving a nonexistent object results in L{NonexistentObject}.
        """
        objectId = u'sha256:NOSUCHOBJECT'
        d = self.contentStore.getObject(objectId)
        return self.assertFailure(d, NonexistentObject
            ).addCallback(lambda e: self.assertEquals(e.objectId, objectId))



class MigrationTests(TestCase):
    """
    Tests for some migration-related stuff.
    """
    def setUp(self):
        self.store = Store(self.mktemp())
        self.contentStore = ContentStore(store=self.store, hash=u'sha256')
        object.__setattr__(self.contentStore, '_deferToThreadPool', execute)
        self.mockStore = MockContentStore(store=self.store)


    def _mkObject(self):
        """
        Inject an object for testing.
        """
        return ImmutableObject(
            store=self.store,
            hash=u'somehash',
            contentDigest=u'quux',
            content=self.store.newFilePath('foo'),
            contentType=u'application/octet-stream')


    def test_migrateTo(self):
        """
        A migration is initialized with the current range of stored objects.
        """
        objs = [self._mkObject() for _ in xrange(5)]

        dest = ContentStore(store=self.store, hash=u'sha256')
        migration = self.contentStore.migrateTo(dest)
        self.assertIdentical(migration.source, self.contentStore)
        self.assertIdentical(migration.destination, dest)
        self.assertEquals(migration.start, 0)
        self.assertEquals(migration.end, objs[-1].storeID)
        self.assertEquals(migration.current, -1)


    def test_migration(self):
        """
        Migration replicates all objects in this store to the destination.
        """
        def _mkObject(content):
            return self.contentStore._storeObject(
                content=content,
                contentType=u'application/octet-stream')

        obj1 = _mkObject(u'object1')
        obj2 = _mkObject(u'object2')

        dest = self.mockStore
        migration = self.contentStore.migrateTo(dest)
        d = migration.run()

        # Already running, so a new run should not be started
        self.assertIdentical(migration.run(), None)

        # This is created after the migration, so should not be migrated
        _mkObject(u'object2')

        def _verify(ign):
            self.assertEquals(
                dest.events,
                [('storeObject', dest, self.successResultOf(obj1.getContent()),
                  obj1.contentType, obj1.metadata, obj1.created, obj1.objectId),
                 ('storeObject', dest, self.successResultOf(obj2.getContent()),
                  obj2.contentType, obj2.metadata, obj2.created,
                  obj2.objectId)])
        return d.addCallback(_verify)


    def test_nextObject(self):
        """
        L{LocalStoreMigration._nextObject} obtains the next object after the
        most recently processed object, and flags it for migration.
        """
        migration = LocalStoreMigration(
            store=self.store,
            start=0,
            current=-1,
            end=1000,
            source=self.contentStore,
            destination=self.contentStore)
        obj1 = self._mkObject()
        obj2 = self._mkObject()
        m1 = migration._nextObject()
        self.assertIdentical(m1.obj, obj1)
        m2 = migration._nextObject()
        self.assertIdentical(m2.obj, obj2)
        m3 = migration._nextObject()
        self.assertIdentical(m3, None)


    def _mkMigrationJunk(self):
        """
        Set up some test state for migrations.
        """
        obj = self.contentStore._storeObject(
            content='foo',
            contentType=u'application/octet-stream')
        migration = LocalStoreMigration(
            store=self.store,
            start=0,
            current=-1,
            end=1000,
            source=self.contentStore,
            destination=self.mockStore)
        pendingMigration = PendingMigration(
            store=self.store,
            parent=migration,
            obj=obj)
        return obj, migration, pendingMigration


    def test_attemptMigrationSucceeds(self):
        """
        When a migration attempt succeeds, the tracking object is deleted.
        """
        obj, migration, pendingMigration = self._mkMigrationJunk()
        def _cb(ign):
            # .store is set to None on deletion
            self.assertIdentical(pendingMigration.store, None)
        return pendingMigration.attemptMigration().addCallback(_cb)


    def test_attemptMigrationFails(self):
        """
        When a migration attempt fails, the tracking object is not deleted, and
        the trackback is stored and logged.
        """
        obj, migration, pendingMigration = self._mkMigrationJunk()

        def _explode(*a, **kw):
            return fail(ValueError('42'))
        object.__setattr__(self.mockStore, 'storeObject', _explode)

        def _eb(f):
            # .store is set to None on deletion
            self.assertNotIdentical(pendingMigration.store, None)
            tb = pendingMigration.lastFailure
            [tb2] = self.flushLoggedErrors(ValueError)
            self.assertIn(u'ValueError: 42', tb)
            self.assertEquals(tb.encode('ascii'), tb2.getTraceback())

        d = pendingMigration.attemptMigration()
        return self.assertFailure(d, ValueError).addErrback(_eb)



@implementer(IContentStore)
class MockContentStore(Item):
    """
    Mock content store that just logs calls.

    @ivar events: A list of logged calls.
    """

    dummy = integer()
    events = inmemory()
    migrationDestination = inmemory()

    def __init__(self, events=None, **kw):
        super(MockContentStore, self).__init__(**kw)
        if events is None:
            self.events = []
        else:
            self.events = events


    # IContentStore

    def getObject(self, objectId):
        self.events.append(('getObject', self, objectId))
        return fail(NonexistentObject(objectId))


    def storeObject(self, content, contentType, metadata={}, created=None,
                    objectId=None):
        self.events.append(
            ('storeObject', self, content, contentType, metadata, created,
             objectId))
        return succeed(u'sha256:FAKE')


    def migrateTo(self, destination):
        self.migrationDestination = destination
        return TestMigration(store=destination.store)



@implementer(IUploadScheduler)
class MockUploadScheduler(object):
    """
    Mock implementation of L{IUploadScheduler}.
    """

    def __init__(self):
        self.uploads = []


    def scheduleUpload(self, objectId, backend):
        self.uploads.append((objectId, backend))



@implementer(IUploadScheduler)
class NullUploadScheduler(object):
    """
    Upload scheduler that does nothing.
    """
    def scheduleUpload(self, objectId, backend):
        pass



class StoreBackendTests(TestCase):
    """
    Tests for content store backend functionality.
    """
    def setUp(self):
        self.store = Store(self.mktemp())
        self.contentStore1 = ContentStore(store=self.store)
        object.__setattr__(self.contentStore1, '_deferToThreadPool', execute)
        self.contentStore1.storeObject(content='somecontent',
                                       contentType=u'application/octet-stream')
        self.testObject = self.store.findUnique(ImmutableObject)

        self.contentStore2 = ContentStore(store=self.store)
        object.__setattr__(self.contentStore2, '_deferToThreadPool', execute)


    def test_getSiblingExists(self):
        """
        Calling getSiblingObject with an object ID that is present in the local
        store just returns the local object.
        """
        d = self.contentStore1.getSiblingObject(self.testObject.objectId)
        def _cb(o):
            self.o = o
        d.addCallback(_cb)
        self.assertIdentical(self.o, self.testObject)


    def _retrievalTest(self):
        o = self.successResultOf(
            self.contentStore2.getSiblingObject(self.testObject.objectId))
        self.assertEquals(
            self.successResultOf(o.getContent()), 'somecontent')
        o2 = self.successResultOf(
            self.contentStore2.getObject(self.testObject.objectId))
        self.assertIdentical(o, o2)


    def test_getSiblingExistsRemote(self):
        """
        Calling getSiblingObject with an object ID that is missing locally, but
        present in one of the sibling stores, will retrieve the object, as well
        as inserting it into the local store.
        """
        self.store.powerUp(self.contentStore1, ISiblingStore)
        self._retrievalTest()


    def test_getSiblingExistsBackend(self):
        """
        If an object is missing in local and sibling stores, but present in a
        backend store, the object will be retrieved from the backend store.
        """
        self.store.powerUp(self.contentStore1, IBackendStore)
        self._retrievalTest()


    def test_siblingBeforeBackend(self):
        """
        When looking for a missing object, sibling stores are tried before
        backend stores.
        """
        events = []

        siblingStore = MockContentStore(store=self.store, events=events)
        self.store.powerUp(siblingStore, ISiblingStore)

        backendStore = MockContentStore(store=self.store, events=events)
        self.store.powerUp(backendStore, IBackendStore)

        def _cb(e):
            self.assertEquals(
                events,
                [('getObject', siblingStore, u'sha256:aoeuaoeu'),
                 ('getObject', backendStore, u'sha256:aoeuaoeu')])
        return self.assertFailure(
            self.contentStore2.getSiblingObject(u'sha256:aoeuaoeu'),
            NonexistentObject).addCallback(_cb)


    def test_getSiblingMissing(self):
        """
        Calling getSiblingObject with an object ID that is missing everywhere
        raises L{NonexistentObject}.
        """
        self.store.powerUp(self.contentStore1, ISiblingStore)
        objectId = u'sha256:NOSUCHOBJECT'
        d = self.contentStore2.getSiblingObject(objectId)
        return self.assertFailure(d, NonexistentObject
            ).addCallback(lambda e: self.assertEquals(e.objectId, objectId))


    def test_storeObject(self):
        """
        Storing an object also causes it to be scheduled for storing in all
        backend stores.
        """
        contentStore = ContentStore(store=self.store)
        backendStore = MockContentStore(store=self.store)
        self.store.powerUp(backendStore, IBackendStore)
        backendStore2 = MockContentStore(store=self.store)
        self.store.powerUp(backendStore2, IBackendStore)
        scheduler = MockUploadScheduler()
        self.store.inMemoryPowerUp(scheduler, IUploadScheduler)

        contentStore.storeObject(content='somecontent',
                                 contentType=u'application/octet-stream')
        testObject = self.store.findUnique(ImmutableObject)
        pu = scheduler.uploads
        self.assertEquals(len(pu), 2)
        self.assertEquals(pu[0][0], testObject.objectId)
        self.assertEquals(pu[1][0], testObject.objectId)
        for objectId, backend in pu:
            if backend is backendStore:
                break
        else:
            self.fail('No pending upload for backendStore')

        for objectId, backend in pu:
            if backend is backendStore2:
                break
        else:
            self.fail('No pending upload for backendStore2')



class _PendingUploadTests(TestCase):
    """
    Tests for L{_PendingUpload}.
    """
    def setUp(self):
        self.store = Store(self.mktemp())
        self.contentStore = ContentStore(store=self.store)
        object.__setattr__(self.contentStore, '_deferToThreadPool', execute)
        self.store.powerUp(self.contentStore, IContentStore)
        self.contentStore.storeObject(content='somecontent',
                                      contentType=u'application/octet-stream')
        self.testObject = self.store.findUnique(ImmutableObject)
        self.backendStore = MockContentStore(store=self.store)
        self.pendingUpload = _PendingUpload(store=self.store,
                                            objectId=self.testObject.objectId,
                                            backend=self.backendStore)


    def test_successfulUpload(self):
        """
        When an upload attempt is made, the object is stored to the backend
        store. If this succeeds, the L{_PendingUpload} item is deleted.
        """
        def _cb(ign):
            self.assertEquals(
                self.backendStore.events,
                [('storeObject',
                  self.backendStore,
                  'somecontent',
                  u'application/octet-stream',
                  {},
                  self.testObject.created,
                  self.testObject.objectId)])
            self.assertRaises(ItemNotFound,
                              self.store.findUnique,
                              _PendingUpload)
        return self.pendingUpload.attemptUpload().addCallback(_cb)


    def test_failedUpload(self):
        """
        When an upload attempt is made, the object is stored to the backend
        store. If this fails, the L{_PendingUpload} item has its scheduled time
        updated.
        """
        def _storeObject(content, contentType, metadata={}, created=None,
                         objectId=None):
            raise ValueError('blah blah')
        object.__setattr__(self.backendStore, 'storeObject', _storeObject)

        nextScheduled = self.pendingUpload.scheduled + timedelta(minutes=5)
        def _nextAttempt():
            return nextScheduled
        object.__setattr__(self.pendingUpload, '_nextAttempt', _nextAttempt)

        self.successResultOf(self.pendingUpload.attemptUpload())
        self.assertIdentical(self.store.findUnique(_PendingUpload),
                             self.pendingUpload)
        self.assertEquals(self.pendingUpload.scheduled,
                          nextScheduled)
        errors = self.flushLoggedErrors(ValueError)
        self.assertEquals(len(errors), 1)



class ObjectCreatorTests(TestCase):
    """
    Tests for L{ObjectCreator}.
    """
    def setUp(self):
        self.store = Store(self.mktemp())
        self.contentStore = ContentStore(store=self.store, hash=u'sha256')
        self.creator = ObjectCreator(self.contentStore)


    def test_correctContentMD5(self):
        """
        Submitting a request with a Content-MD5 header that agrees with the
        uploaded data should succeed.
        """
        req = FakeRequest()
        req.received_headers['content-md5'] = '72VMQKtPF0f8aZkV1PcJAg=='
        req.content = StringIO('testdata')
        return self.creator.handlePUT(req)


    def test_incorrectContentMD5(self):
        """
        Submitting a request with a Content-MD5 header that disagrees with the
        uploaded data should fail.
        """
        req = FakeRequest()
        req.received_headers['content-md5'] = '72VMQKtPF0f8aZkV1PcJAg=='
        req.content = StringIO('wrongdata')
        self.assertRaises(ValueError, self.creator.handlePUT, req)


    def test_missingContentMD5(self):
        """
        Submitting a request with no Content-MD5 header should succeed.
        """
        req = FakeRequest()
        req.content = StringIO('wrongdata')
        return self.creator.handlePUT(req)



class ImmutableObjectTests(TestCase):
    """
    Tests for L{ImmutableObject}.
    """
    def setUp(self):
        self.store = Store(self.mktemp())
        self.contentStore = ContentStore(store=self.store)
        object.__setattr__(self.contentStore, '_deferToThreadPool', execute)
        self.contentStore.storeObject(content='somecontent',
                                      contentType=u'application/octet-stream')
        self.testObject = self.store.findUnique(ImmutableObject)


    def test_metadata(self):
        """
        Metadata is not yet supported, so L{ImmutableObject.metadata} should be
        an empty dict.
        """
        self.assertEquals(self.testObject.metadata, {})


    def test_verify(self):
        """
        Verification should succeed when the object contents has not been
        altered.
        """
        self.testObject.verify()


    def test_verifyDamaged(self):
        """
        Verification should fail if the object contents is modified.
        """
        self.testObject.content.setContent('garbage!')
        self.assertRaises(CorruptObject, self.testObject.verify)


    def test_getContent(self):
        """
        L{ImmutableObject.getContent} returns the contents of the object.
        """
        self.assertEquals(
            self.successResultOf(self.testObject.getContent()), 'somecontent')


    def test_objectId(self):
        """
        The object ID is composed of the digest function and content digest,
        separated by a colon.
        """
        self.assertEquals(
            self.testObject.objectId,
            'sha256:d5a3477d91583e65a7aba6f6db7a53e2de739bc7bf8f4a08f0df0457b637f1fb')


    def test_adaptToResource(self):
        """
        Adapting L{ImmutableObject} to L{IResource} gives us a L{File} instance
        pointing at the on-disk blob.
        """
        res = IResource(self.testObject)
        self.assertIsInstance(res, File)
        self.assertEquals(res.fp, self.testObject.content)
        self.assertEquals(res.type, 'application/octet-stream')
        self.assertEquals(res.encoding, None)



@implementer(IMigration)
class TestMigration(Item):
    """
    Test double implementing IMigration.
    """
    powerupInterfaces = [IMigration]

    ran = integer(default=0)

    def run(self):
        self.ran += 1



class MigrationManagerTests(TestCase):
    """
    Tests for L{MigrationManager}.
    """
    def setUp(self):
        self.store = Store()
        self.manager = MigrationManager(store=self.store)


    def test_installService(self):
        """
        The service is started when it is installed into a running store, and
        stopped when it is deleted.
        """
        IService(self.store).startService()
        installOn(self.manager, self.store)
        self.assertTrue(self.manager.running)
        self.manager.deleteFromStore()
        self.assertFalse(self.manager.running)


    def test_serviceRunsMigrations(self):
        """
        Starting the service runs all existing migrations.
        """
        m1 = TestMigration(store=self.store)
        m2 = TestMigration(store=self.store)
        self.store.powerUp(m1)
        self.store.powerUp(m2)
        self.assertEquals(m1.ran, 0)
        self.assertEquals(m2.ran, 0)
        self.manager.startService()
        self.assertEquals(m1.ran, 1)
        self.assertEquals(m2.ran, 1)


    def test_startMigration(self):
        """
        Starting a migration invokes the implementation on the source store.
        """
        source = MockContentStore()
        destination = MockContentStore(store=self.store)
        result = self.manager.migrate(source, destination)
        self.assertEquals(result.ran, 1)
        self.assertEquals(source.migrationDestination, destination)
        self.assertEquals(IMigration(self.store), result)



class ProxyStore(proxyForInterface(IContentStore), Item):
    """
    Content store that just proxies to some other object.

    Usually used with a L{ContentStore} in another store.
    """
    dummy = integer()
    original = inmemory()

    def __init__(self, store, original):
        super(ProxyStore, self).__init__(store=store)
        self.original = original



class InsaneStore(Item):
    """
    Content store that doesn't return the object you asked for.
    """
    dummy = integer()

    def getObject(self, *args, **kwargs):
        return succeed(MemoryObject(
            content=u'haha',
            hash=u'sha256',
            contentDigest=u'what is a digest even',
            contentType=u'insanity/madness',
            created=Time()))



class VerificationTests(TestCase):
    """
    Tests for integrity verification.
    """
    def _store(self):
        store = Store(self.mktemp())
        store.inMemoryPowerUp(NullUploadScheduler(), IUploadScheduler)
        contentStore = ContentStore(store=store)
        store.powerUp(contentStore, IContentStore)
        object.__setattr__(contentStore, '_deferToThreadPool', execute)
        return contentStore


    def _storeObject(self, contentStore, content, contentType):
        return self.successResultOf(
            contentStore.storeObject(
                content=content, contentType=contentType).addCallback(
                    contentStore.getObject))


    def _verify(self, contentStore, obj):
        """
        Run a verification on an object.
        """
        migration = LocalStoreMigration(
            store=obj.store, source=contentStore, destination=None, start=0,
            current=0, end=0)
        pending = PendingMigration(store=obj.store, parent=migration, obj=obj)
        return pending._verify()


    def test_oneStoreIntact(self):
        """
        Verifying an object with the correct digest, and no other backends,
        completes without error.
        """
        contentStore = self._store()
        obj = self._storeObject(
            contentStore=contentStore,
            content='somecontent',
            contentType=u'application/octet-stream')
        self.successResultOf(self._verify(contentStore, obj))


    def test_oneStoreDamaged(self):
        """
        Verifying an object with incorrect content, and no other backends,
        results in a L{NoGoodCopies} failure.
        """
        contentStore = self._store()
        obj = self._storeObject(
            contentStore=contentStore,
            content='somecontent',
            contentType=u'application/octet-stream')
        obj.content.setContent('damaged')
        f = self.failureResultOf(self._verify(contentStore, obj), NoGoodCopies)
        self.assertEqual(f.value.objectId, obj.objectId)


    def test_twoStoresIntact(self):
        """
        Verifying an object with the correct content, and one backend with a
        copy that also has the correct content, completes without error.
        """
        contentStore = self._store()
        store = contentStore.store
        contentStore2 = self._store()
        obj = self._storeObject(
            contentStore=contentStore,
            content='somecontent',
            contentType=u'application/octet-stream')
        obj2 = self._storeObject(
            contentStore=contentStore2,
            content='somecontent',
            contentType=u'application/octet-stream')
        store.inMemoryPowerUp(contentStore2, IBackendStore)
        self.successResultOf(self._verify(contentStore, obj))


    def test_twoStoresRepair(self):
        """
        Verifying an object with the correct content, and one backend with a
        copy that has incorrect content, repairs the object in the backend
        successfully.
        """
        contentStore = self._store()
        store = contentStore.store
        contentStore2 = self._store()
        obj = self._storeObject(
            contentStore=contentStore,
            content='somecontent',
            contentType=u'application/octet-stream')
        obj2 = self._storeObject(
            contentStore=contentStore2,
            content='somecontent',
            contentType=u'application/octet-stream')
        obj2.content.setContent('damaged')
        store.inMemoryPowerUp(contentStore2, IBackendStore)
        self.successResultOf(self._verify(contentStore, obj))
        self.assertEquals(obj2.content.getContent(), 'somecontent')


    def test_twoStoresRepairFromRemote(self):
        """
        Verifying an object with incorrect content, but one backend with a
        copy that has the correct content, repairs the local object
        successfully.
        """
        contentStore = self._store()
        store = contentStore.store
        contentStore2 = self._store()
        obj = self._storeObject(
            contentStore=contentStore,
            content='somecontent',
            contentType=u'application/octet-stream')
        obj2 = self._storeObject(
            contentStore=contentStore2,
            content='somecontent',
            contentType=u'application/octet-stream')
        obj.content.setContent('damaged')
        store.inMemoryPowerUp(contentStore2, IBackendStore)
        self.successResultOf(self._verify(contentStore, obj))
        self.assertEquals(obj.content.getContent(), 'somecontent')


    def test_twoStoresBothDamaged(self):
        """
        Verifying an object with incorrect content, and all backends with
        corrupt copies, results in a L{NoGoodCopies} failure.
        """
        contentStore = self._store()
        store = contentStore.store
        contentStore2 = self._store()
        obj = self._storeObject(
            contentStore=contentStore,
            content='somecontent',
            contentType=u'application/octet-stream')
        obj2 = self._storeObject(
            contentStore=contentStore2,
            content='somecontent',
            contentType=u'application/octet-stream')
        obj.content.setContent('damaged')
        obj2.content.setContent('also damaged')
        store.inMemoryPowerUp(contentStore2, IBackendStore)
        f = self.failureResultOf(self._verify(contentStore, obj), NoGoodCopies)
        self.assertEqual(f.value.objectId, obj.objectId)


    def test_twoStoresMissing(self):
        """
        Verifying an object with the correct content, and one backend with the
        object missing, uploads the object to the backend.
        """
        contentStore = self._store()
        store = contentStore.store
        obj = self._storeObject(
            contentStore=contentStore,
            content='somecontent',
            contentType=u'application/octet-stream')
        contentStore2 = self._store()
        store.inMemoryPowerUp(contentStore2, IBackendStore)
        self.successResultOf(self._verify(contentStore, obj))
        obj2 = self.successResultOf(contentStore2.getObject(obj.objectId))
        self.assertEquals(obj2.content.getContent(), 'somecontent')


    def test_misbehavingBackend(self):
        """
        Verifying an object with a backend that returns the wrong object
        results in an L{UnexpectedDigest} failure.
        """
        contentStore = self._store()
        store = contentStore.store
        obj = self._storeObject(
            contentStore=contentStore,
            content='somecontent',
            contentType=u'application/octet-stream')
        contentStore2 = InsaneStore(store=store)
        store.inMemoryPowerUp(contentStore2, IBackendStore)
        f = self.failureResultOf(
            self._verify(contentStore, obj), UnexpectedDigest)
        self.assertEqual(f.value.objectId, obj.objectId)
